import numpy as np
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import skimage as sk
import matplotlib as mpl
mpl.rcParams['image.cmap'] = 'gray'

from scipy.interpolate import interp1d

def arclength_param(line) :
    "Arclength parametrisation of a piecewise affine curve."
    vel = line[1:, :] - line[:-1, :]
    vel = np.sqrt(np.sum( vel ** 2, 1 ))
    return np.hstack( ( [0], np.cumsum( vel, 0 ) ) )
def arclength(line) :
    "Total length of a piecewise affine curve."
    return arclength_param(line)[-1]
    
def resample(line, npoints) :
    "Resamples a curve by arclength through linear interpolation."
    s = arclength_param(line)
    f = interp1d(s, line, kind = 'linear', axis = 0, assume_sorted = True)
    p = f( np.linspace(0, s[-1], npoints) )
    if np.array_equal(p[0], p[-1]) : # i.e. p is a loop
        p = p[:-1]
    return p


def process_contour( c, N ) : # N.B.: c[0] == c[-1]
    "Turns a circular contour line into a "
    f = np.vstack( ( [c[-2]], c) )
    u = np.diff(f, n=1, axis=0)
    v = .5*( u[1:,:] + u[:-1,:] )
    a = np.diff(f, n=2, axis=0)
    k = np.abs( a[:,0]*v[:,1] - a[:,1]*v[:,0] ) / (np.sum( v**2, 1))**(3/2) # curvature
    imax = np.argmax(k) # index of maximum curvature
    
    L = np.sum( np.sqrt(np.sum( u**2, 1))) # total length of the curve
    C = np.vstack( (c[imax:-1,:], c) ) # whole loop
    U = np.diff( C, n=1, axis=0)
    S = np.sqrt(np.sum(U**2, 1))
    Ls = np.hstack( ( [0], np.cumsum( S, 0 ) ) ) # arclength parametrization from imax
    C1 = C[Ls <= L/2]      # Take the first half of the contour line, starting from imax
    C2 = C[(L/2 <= Ls) & (Ls <= L)][::-1] # Second half
    return .5*( resample(C1, N) + resample(C2,N))
    
from skimage.util import invert
from skimage.filters import sobel
from scipy import ndimage as ndi
from skimage.morphology import medial_axis, skeletonize, skeletonize_3d
from skimage.morphology import opening, thin, binary_opening, binary_closing, binary_erosion
from skimage import measure
from skimage.filters import gaussian

def extract_contours(img, display=False, 
                     gradient_threshold = .1, skeleton_blur = 1., skeleton_threshold=.1,
                     erode=3, min_length=100) :
    "Segment an image of worm using simple morphological operations."
    i_img = 1- img #invert( np.mean(img, 2) )
    norm_grad = sobel(i_img)
    edges     = binary_closing( norm_grad > gradient_threshold )
    fill_worm = binary_opening( ndi.binary_fill_holes(edges) )
    eroded_worm = binary_erosion(fill_worm, selem = [[True,]*erode,]*erode)
    
    skeleton  = skeletonize_3d(eroded_worm).astype(float)/255. # in practice, skeletonize_3d produces less branches
    blurred_skeleton = gaussian( skeleton, skeleton_blur)
    contours  = measure.find_contours(blurred_skeleton, skeleton_threshold)
    contours  = [con for con in contours if len(con) > min_length ]
    
    if contours == [] :
        contours = extract_contours(img, gradient_threshold = gradient_threshold-0.01)
    
    if display :
        fig, axes = plt.subplots(nrows=3, ncols=2, figsize=(8, 8),
                                 sharex=True, sharey=True)

        ax = axes.ravel()

        ax[0].imshow(i_img)
        ax[0].set_title('original', fontsize=20)
        ax[1].imshow(norm_grad)
        ax[1].set_title('norm_grad', fontsize=20)
        ax[2].imshow(edges)
        ax[2].set_title('edges', fontsize=20)
        ax[3].imshow(fill_worm)
        ax[3].set_title('segmentation', fontsize=20)
        ax[4].imshow(eroded_worm)
        ax[4].set_title('eroded segmentation', fontsize=20)
        ax[5].imshow(skeleton)
        ax[5].set_title('skeleton', fontsize=20)
        
        fig.tight_layout()

        plt.figure(figsize=(8,8))
        plt.imshow(blurred_skeleton)
        plt.title('blurred skeleton', fontsize=20)
        plt.show()
        
        plt.figure(figsize=(8,8))

        plt.imshow(img)
        for n, contour in enumerate(contours):
            plt.plot(contour[:, 1], contour[:, 0], linewidth=2)
        plt.axis('image')
        plt.show()

    return contours

def reorient_worm( Pt ) :
    Pt = np.stack(Pt) # Shape (T,N,2)
    # First, make sure that everyone's on the same wavelength
    for t, P in enumerate(Pt) :
        if t > 0 :
            Head = Pt[t-1, 0, :]
            Tail = Pt[t-1,-1,:]
            H2H  = np.sum( (P[0]  - Head)**2)
            H2T  = np.sum( (P[-1] - Head)**2)
            T2H  = np.sum( (P[0]  - Tail)**2)
            T2T  = np.sum( (P[-1] - Tail)**2)
            
            if H2H + T2T > H2T + T2H :
                Pt[t,:,:] = Pt[t,::-1,:]
                
    # Then, make sure that the Head is in first position
    dt = (Pt[1] - Pt[0])[1:]
    ds = Pt[0,1:] - Pt[0,:-1]
    if np.sum(ds*dt) > 0 :
        Pt[:,:,:] = Pt[:,::-1,:]
    return Pt

def process_movie( imgs, N=20 ) :
    
    # Pi_t.shape = (T,K,N,2)
    Pi_t = np.array( [ [process_contour( con, N ) for con in extract_contours(img)] for img in imgs ] )
    Ni_t = np.array([len(Pi) for Pi in Pi_t])
    K    = Ni_t[0]
    if np.all([ nt == K for nt in Ni_t]) : print("Detected {} worms.".format(K))
    else : 
        print("Iconsistent numbers of detected worms.")
        print(Ni_t)
        print("Faulty indices : ", np.where(Ni_t != K))
        #raise ValueError("Discard the frames where a worm touches the boundary.")
        print("We only keep the longest contour.")
        K = 1
        #Pi_t = [ P for (P,l) in zip(Pi_t,Ni_t) if l > 0 ]
        contour_lengths = [ [arclength(P) for P in Pi] for Pi in Pi_t ]
        max_index = [ np.argmax(cl) for cl in contour_lengths ]
        Pi_t = np.array( [[Pi[ind]] for (ind, Pi) in zip(max_index, Pi_t)] )
        
    # Separate the different trajectories
    Centers_t = np.copy( Pi_t[:,:,N//2,:] )
    Trajectories = [ [ Pi_0 for Pi_0 in Pi_t[0] ] ]
    Centers   = Centers_t[0,:,:]
    for (Pi_s, C_s) in zip( Pi_t[1:], Centers_t[1:] ) : # Shapes (K,N,2) and (K,2)
        newCenters = []
        newWorms   = []
        for k in range(K) :
            dist_to_Ck = np.sum( ( C_s - Centers[k])**2, 1 )
            i_k = np.argmin( dist_to_Ck )
            newCenters.append(np.copy( C_s[i_k]) )
            newWorms.append(Pi_s[i_k])
            C_s[i_k,:] = np.inf
            
        Centers = newCenters
        Trajectories.append( newWorms )
        
    Trajectories = list(map(list, zip(*Trajectories))) # Transpose
    return [reorient_worm(Pt) for Pt in Trajectories]

import os

def segment_worms(day=5, N=20, screenshot=False) :
    if not os.path.exists('data/day_{}'.format(day)):
        os.makedirs('data/day_{}'.format(day))
        
    Imgs = [ mpimg.imread('data/worms/day_{}/{}.png'.format(day, ind)) for ind in range(1, 201)]
    Pt_s = process_movie(Imgs, N=N)
    
    for s, Pt in enumerate( Pt_s ):
        Zt = Pt[:,:,0] + 1j*Pt[:,:,1]
        Pt = np.stack((Pt[:,:,1],Pt[:,:,0]),axis = 2)
        AA = Pt.reshape(Pt.shape[0], Pt.shape[1]*2)
        np.savetxt('data/day_{}.csv'.format(day), AA, delimiter = ',')

    if screenshot :
        Ps_t = list(map(list, zip(*Pt_s))) # Transpose
        plt.figure()
        for t, (Img, Ps) in enumerate( zip(Imgs, Ps_t)) :
            plt.clf()
            plt.imshow(Img)
            colors = ['red', 'green', 'blue', 'magenta', 'pink', 'coral'][:len(Ps)]
            for (col,P) in zip(colors,Ps) : 
                plt.scatter(P[0,1],P[0,0], 9*25, color=col)
                plt.scatter(P[:,1],P[:,0], color=col)
        
            plt.tight_layout()
            plt.savefig('data/day_{}/worm_{}.png'.format(day, t))
            #plt.show()


